%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Proposed Algorithm: Entropy Enhanced Covariance Matrix Adaptation Evolution Strategy (EE_ CMAES)
%Dt: 09/02/2018
% Developers: Kartik S. Pandya, Electrical Engg. Dept., CSPIT, CHARUSAT,
% INDIA. (email:kartikpandya.ee@charusat.ac.in)
% Jigar Sarda,Electrical Engg. Dept., CSPIT, CHARUSAT,
% INDIA. email: jigarsarda.ee@charusat.ac.in
% Test bed A: Stochastic OPF in presence of renewable energy and
% controllable loads. 2018 Grid optimization competition
% Publically and freely available MATLAB codes of Cross Entropy method (website: http://sites.ieee.org/psace-mho/2017-smart-grid-operation-problems-competition-panel/)
%and CMAES method (website: http://www.lri.fr/~hansen/purecmaes.m) are used 
%and modified to include them into OPF.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function EE_CMAES(fhd,ii,jj,kk,args)
% Paste and adapt the following
% parts to your implementation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%addpath ('matpower6.0b1')
global proc
global ps
VRmin=[ps.x_min  50 300  0  0]; % PLEASE NOTE THAT 4 LOWER BOUNDS OF CONTROLABLE LOADS WERE ADDED REGARDING 2017 TEST BED 1
    % Individuals' upper bounds.
    VRmax=[ps.x_max 158 420 53 53]; 
    Xmin=VRmin; % 1*35

% Particles' upper bounds
Xmax=VRmax;
define_constants;
% Reinitialization of local
% function evaluation counter.
proc.i_eval=0;
D=ps.D+4;
%gbestval=1e10
% Function evaluation at which
% the last update of the global
% best solution occurred.
% Refers to the internal evaluation
% using static penalty constraint
% handling method.
proc.last_improvement=1;
% Signalizing your
% to stop running.
proc.finish=0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CE Initializations
n_samples = proc.pop_size;
ro = 0.1;
sigma = 0.85;
beta = 0.9;
q = 5;
indexRo = floor( ( 1 - ro ) * n_samples );
% indexRo = n_samples - 10;
H = zeros( 1, n_samples );
pos = zeros( n_samples, ps.D+4 );
gammaIter = zeros( 1, 1e6 );
x_min = zeros( 1, ps.D+4 );
x_max = zeros( 1, ps.D+4 );
n_trials = ones( 1, ps.D+4 );
meanPar = zeros( 1, ps.D+4 );
stdPar = zeros( 1, ps.D+4 );
for i = 1 : ps.D+4
    if i <= ps.D_cont || i>=32% Continuous variables
        x_min( 1, i ) = Xmin( 1, i );
        x_max( 1, i ) = Xmax( 1, i );
        meanPar( 1, i ) = x_min( 1, i ) + rand() * ( x_max( 1, i ) - x_min( 1, i ) );
        stdPar( 1, i ) = ( x_max( 1, i ) - x_min( 1, i ) ) * 5;
    elseif i <= ps.D_cont + ps.n_OLTC % OLTC
        x_min( 1, i ) = Xmin( 1, i ) - 0.499;
        x_max( 1, i ) = Xmax( 1, i ) + 0.499;
        meanPar( 1, i ) = x_min( 1, i ) + rand() * ( x_max( 1, i ) - x_min( 1, i ) );
        stdPar( 1, i ) = ( x_max( 1, i ) - x_min( 1, i ) ) * 5;
    else % Shunt compensators
        x_min( 1, i ) = 0;
        x_max( 1, i ) = 1;
        meanPar( 1, i ) = rand();
    end
end
iter = 0;
memMeanPar( iter + 1, : ) = meanPar;
memStdPar( iter + 1, : ) = stdPar;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Fitness Function Initializations
global ff
ff.coeff = ones( 1, 4 );
ff.weights = zeros( n_samples, 4 );
ff.i_sample = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%
%while 1
    %%%%%%%
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Run CE
    while proc.i_eval <= 15000
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Generate population
        for i = 1 : ps.D+4
            if i <= ps.D_cont + ps.n_OLTC || i>=32 % Continuous variables + OLTC
                tmp_stat = normrnd( meanPar( i ), stdPar( i ), 1, n_samples );
                tmp_stat( tmp_stat < x_min( 1, i ) ) = x_min( 1, i );
                tmp_stat( tmp_stat > x_max( 1, i ) ) = x_max( 1, i );
                tmp_eval = tmp_stat;
            else % Shunt compensators
                tmp_stat = binornd( n_trials( i ), meanPar( i ), 1, n_samples );
                tmp_eval = tmp_stat + x_min( 1, i );
            end
            pos( :, i ) = tmp_eval;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Evaluate population
        [ fit,~,~,pos,~ ] = feval( fhd,ii,jj,kk,args,pos );
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Performs CE correction
        
        % Sorts states according to fitness
        [ Saux, Iaux ] = sort( fit, 2, 'descend' );
        gbestval=min(fit);
        H = zeros( 1, n_samples );
        for i = 1 : n_samples
            if i <= indexRo
                H( Iaux( i ) ) = 0;
            else
                H( Iaux( i ) ) = 1;
            end
        end
        
        % Computes gammaIter
        gamma = Saux( 1, indexRo );
        gammaIter( 1, iter + 1 ) = gamma;
        
        % Applies CE correction
        den = sum( H );
        for i = 1 : ps.D+4
            num = H * pos( :, i );
            tmpMean = 0;
            if den > 0
                tmpMean = ( num / den ) * ( 1 / n_trials( 1, i ) );
            end
            num = H * ( ( pos( :, i ) - tmpMean ) .* ( pos( :, i ) - tmpMean ) );
            tmpStd = 0;
            if den > 0
                tmpStd = sqrt( num / den ) * ( 1 / n_trials( 1, i ) );
            end
            meanPar( i ) = tmpMean * sigma + meanPar( i ) * ( 1 - sigma );
            
            % Update smoothing parameters
            sigmaStd = beta - beta * ( 1 - 1 / ( iter + 1 ) )^q;
            stdPar( i ) = tmpStd * sigmaStd + stdPar( i ) * ( 1 - sigmaStd );
            
            % Update sampling distribution parameters
            if i <= ps.n_gen_PG ||i>=32
                meanPar( i ) = round( meanPar( i ), 2 );
            elseif i <= ps.D_cont
                meanPar( i ) = round( meanPar( i ), 4 );
            elseif i <= ps.D_cont + ps.n_OLTC
                % meanPar( i ) = round( meanPar( i ), 2 );
            else
                if meanPar( i ) < 0.01
                    meanPar( i ) = 0.01;
                elseif meanPar( i ) > 0.99
                    meanPar( i ) = 0.99;
                end
            end
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Update the weights of the fitness function
        tmpMat = ff.weights( 1:ff.i_sample, : );
        tmp = max( tmpMat );
        ff.coeff = 10 .^ round( log10( max( tmp )./ tmp )  );
        ff.coeff( isinf( ff.coeff ) ) = 1;
        ff.i_sample = 0;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Update CE iteration
        iter = iter + 1;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        memMeanPar( iter + 1, : ) = meanPar;
        memStdPar( iter + 1, : ) = stdPar;
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Run CMAES
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     VRmin=[ps.x_min  50 300  0  0]; % PLEASE NOTE THAT 4 LOWER BOUNDS OF CONTROLABLE LOADS WERE ADDED REGARDING 2017 TEST BED 1
    % Individuals' upper bounds.
    VRmax=[ps.x_max 158 420 53 53]; % PLEASE NOTE THAT 4 UPPER BOUNDS OF CONTROLABLE LOADS WERE ADDED REGARDING 2017 TEST BED 1
%pause;
    % % Particles' lower bounds. original 2017
% Xmin=ps.x_min;
% % Particles' upper bounds.
% Xmax=ps.x_max;
% Particles' lower bounds.
Xmin=VRmin;

% Particles' upper bounds
Xmax=VRmax;
pos1=pos'; % D*N
xmean=pos1(:,Iaux(end)); %gbest

 % xmean = rand(D,1);    % objective variables initial point
  %sigma = 0.3;          % coordinate wise standard deviation (step size)
  sigma = 0.001;
   stopfitness = 1e-10;  % stop if fitness < stopfitness (minimization)
   stopeval = 1e3*D^2;   % stop after stopeval number of function evaluations
  
  % Strategy parameter setting: Selection  
  %lambda = 4+floor(3*log(D));  % population size, offspring number
  lambda=proc.pop_size;
  mu = lambda/2;               % number of parents/points for recombination
  weights = log(mu+1/2)-log(1:mu)'; % muXone array for weighted recombination
  mu = floor(mu);        
  weights = weights/sum(weights);     % normalize recombination weights array
  mueff=sum(weights)^2/sum(weights.^2); % variance-effectiveness of sum w_i x_i

  % Strategy parameter setting: Adaptation
  cc = (4+mueff/D) / (D+4 + 2*mueff/D);  % time constant for cumulation for C
  cs = (mueff+2) / (D+mueff+5);  % t-const for cumulation for sigma control
  c1 = 2 / ((D+1.3)^2+mueff);    % learning rate for rank-one update of C
  cmu = min(1-c1, 2 * (mueff-2+1/mueff) / ((D+2)^2+mueff));  % and for rank-mu update
  damps = 1 + 2*max(0, sqrt((mueff-1)/(D+1))-1) + cs; % damping for sigma 
                                                      % usually close to 1
  % Initialize dynamic (internal) strategy parameters and constants
  pc = zeros(D,1); psc = zeros(D,1);   % evolution paths for C and sigma
  B = eye(D,D);                       % B defines the coordinate system diagonal 1 1 1
  F = ones(D,1);                      % diagonal D defines the scaling D *1 : 1 1 
  C = B * diag(F.^2) * B';            % covariance matrix C
  invsqrtC = B * diag(F.^-1) * B';    % C^-1/2 
  eigeneval = 0;                      % track update of B and D
  chiN=D^0.5*(1-1/(4*D)+1/(21*D^2));  % expectation of 
                                      %   ||N(0,I)|| == norm(randn(N,1)) 
  % -------------------- Generation Loop --------------------------------
  % fit2=zeros(1,lambda)
  % arx=zeros(proc.pop_size, D);
  %aaa=10000*(ones(proc.pop_size-lambda , D));
%  fit=zeros(1,lambda);
%  arx_trans=zeros(lambda, D);
  counteval = 0;  % the next 40 lines contain the 20 lines of interesting code 
 % old_arfitness=10000000;

 % while  1% counteval < stopeval
   while proc.i_eval>15000
       
         if proc.i_eval>20000
              sigma = 0.0001;
         end
      % Generate and evaluate lambda offspring
      for k=1:lambda  %pop size
          
          arx(:,k) = xmean + sigma * B * (F .* randn(D,1)); % m + sig * Normal(0,C), arx=35*100
          %arx(:,k) = xmean + sigma * B * (F .* ccpos); % m + sig * Normal(0,C), arx=35*100
          counteval = counteval+1;
      end
      
     
   arx_trans1=arx'; % 100*35, N*D
   
   %% check the limits of variables and fix them.
   for i=1:lambda
       for j=1:D
           if arx_trans1(i,j)<Xmin( j )
               arx_trans1(i,j)=Xmin( j );    
           elseif arx_trans1(i,j)>Xmax( j )
              arx_trans1(i,j)=Xmax( j );  
           end
       end
   end
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
   % round the 14-31 variables
   for i=1:lambda
       for j=14:31
            arx_trans1(i,j)=round(arx_trans1(i,j));
       end
   end
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   arx_trans=arx_trans1; %size proc.pop_size*D, N*D
  
   
    [arfitness, ~, ~,arx_trans, ~ ] = feval(fhd,ii,jj,kk,args,arx_trans);
    
     
    
      % Sort by fitness and compute weighted mean into xmean
      [arfitness, arindex] = sort(arfitness); % minimization
     
      %%%%%%%%%%%%%%%%%%%
       % UPDATE GLOBAL BEST
    [ tmpgbestval, gbestid ] = min(arfitness  );
    if tmpgbestval < gbestval
        gbestval = tmpgbestval;
        gbest =arx_trans ( gbestid, : );
    end
      
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Update the weights of the fitness function
        tmpMat = ff.weights( 1:ff.i_sample, : );
        tmp = max( tmpMat );
        ff.coeff = 10 .^ round( log10( max( tmp )./ tmp )  );
        ff.coeff( isinf( ff.coeff ) ) = 1;
        ff.i_sample = 0;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      
      %%%%%%%%%%%%%%%%%%%%%%
     % gbestval=min(arfitness);
%       if arfitness(1)<old_arfitness
%           old_arfitness=arfitness(1);
%       end
      xold = xmean;
      xmean = arx(:,arindex(1:mu))*weights;   % recombination, new mean value
    
      % Cumulation: Update evolution paths
      psc = (1-cs)*psc ... 
            + sqrt(cs*(2-cs)*mueff) * invsqrtC * (xmean-xold) / sigma; 
      hsig = norm(psc)/sqrt(1-(1-cs)^(2*counteval/lambda))/chiN < 1.4 + 2/(D+1);
      pc = (1-cc)*pc ...
            + hsig * sqrt(cc*(2-cc)*mueff) * (xmean-xold) / sigma;

      % Adapt covariance matrix C
      artmp = (1/sigma) * (arx(:,arindex(1:mu))-repmat(xold,1,mu));
      C = (1-c1-cmu) * C ...                  % regard old matrix  
           + c1 * (pc*pc' ...                 % plus rank one update
                   + (1-hsig) * cc*(2-cc) * C) ... % minor correction if hsig==0
           + cmu * artmp * diag(weights) * artmp'; % plus rank mu update

      % Adapt step size sigma
      sigma = sigma * exp((cs/damps)*(norm(psc)/chiN - 1)); 
    
      % Decomposition of C into B*diag(D.^2)*B' (diagonalization)
      if counteval - eigeneval > lambda/(c1+cmu)/D/10  % to achieve O(N^2)
          eigeneval = counteval;
          C = triu(C) + triu(C,1)'; % enforce symmetry
          [B,F] = eig(C);           % eigen decomposition, B==normalized eigenvectors
          F = sqrt(diag(F));        % D is a vector of standard deviations now
          invsqrtC = B * diag(F.^-1) * B';
      end
    
      % Break, if fitness is good enough or condition exceeds 1e14, better termination methods are advisable 
       if arfitness(1) <= stopfitness || max(F) > 1e7 * min(F)
           break;
       end
%%%%%%%%%%
%gbestval= arfitness(1);
%old_arfitness;
% gbest = arx_trans( arindex(1), :); % Return best point of last iteration.
                             % Notice that xmean is expected to be even
                             % better. gbest
  
  proc.i_eval=proc.i_eval+1;
  
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     if proc.finish
         return;
    end
    %%%%%%%%%%%%%%
  end 

  end
    

